from bulk_edit import tasks, log, TaskExecutor
# TODO we need to import all from comunity in order to load class by string from globals()
#      (not nice but working)
from bulk_edit.tasks.community import *
from bulk_edit.tasks.task_base import Task, print_subclass
from selenium import webdriver
from selenium.webdriver.firefox.options import Options
import argparse
import logging
from typing import Dict, List
logger = logging.getLogger(__name__)


def get_input_from_cli():
    parser = argparse.ArgumentParser(prog='bulk_edit',
                                     description='Doing bulk edit jobs for vereinsflieger.de')

    parser.add_argument('action', choices=['list', 'run'],
                        help='Action to perform')
    parser.add_argument('task_name', default='*',
                        help='task name to run / list. Use * to list all task\'s.')
    parser.add_argument('task_parameter', nargs='*', default=[],
                        help='parameter for task (syntax: name=value)')
    parser.add_argument('--uidFile', type=str, dest='uid_file_name',
                        help='If specified the task will be repeated with parameter `uid`s read '
                             'from file (one uid per line)')
    parser.add_argument('-v', '--silent', action='store_true', default=False)
    parser.add_argument('-o', '--keepOpen', action='store_true', default=False)
    parser.add_argument('--credentialFile', type=str,
                        help=('A simple text file with content the following content.\n'
                              'first line: user name\n'
                              'second line: password'))

    options = parser.parse_args()

    return options


if __name__ == '__main__':
    driver = None
    batchProcess = None
    closeDriver = True
    try:
        log.load_config()
        options = get_input_from_cli()
        closeDriver = not options.keepOpen

        if options.action == 'list':
            print('\nAvailable tasks:\n')
            print_subclass(Task, options.task_name)
            exit(0)

        def parse_parameters(params: List[str]) -> dict:
            parameters: Dict[str, str] = {}

            for param in params:
                try:
                    tokens = str(param).split('=')
                    parameters[tokens[0]] = tokens[1]
                except Exception as ex:
                    raise Exception(f"Unexpected parameter format '{param}'. It "
                                    "might be 'name=value'.") from ex
            return parameters

        parameters = parse_parameters(options.task_parameter)

        def parse_className(className: str) -> type:
            try:
                return globals()[className]
            except Exception as ex:
                raise Exception(f"Unexpected Task '{className}'. Call action `list` to show "
                                "available Tasks.") from ex

        task_t = parse_className(options.task_name)

        def generate_webdriver(headless: bool) -> webdriver:
            profile = webdriver.FirefoxProfile()
            options = Options()
            # this line starts FF in headless mode
            options.headless = headless
            return webdriver.Firefox(firefox_profile=profile, options=options)

        driver = generate_webdriver(options.silent)

        # read uid from user input or from file
        taskList = []
        try:
            if options.uid_file_name:
                # check if uid is specified explicit
                if 'uid' in parameters:
                    logger.warning("Option `--uidFile` and `uid` is specified. `uid` will be "
                                "discarded. Instead uids from file will be read.")
                    del parameters['uid']

                with open(options.uid_file_name, "r") as f:
                    line = f.readline()
                    while line:
                        # prepare task
                        task = task_t(driver=driver, uid=int(line), **parameters)
                        taskList.append(task)
                        # makes this loop a for each line
                        line = f.readline()

            else:
                task = task_t(driver=driver, **parameters)
                taskList.append(task)
        except TypeError as ex:
            logger.error(f"Task creation failed with message '{ex}'.\n"
                         "Check if you specified all parameters!")
            exit(1)

        # execute list
        loginParams = {}
        if options.credentialFile:
            loginParams['credentialFile'] = options.credentialFile
        login = tasks.LoginToVf(driver, **loginParams)
        login.execute()

        batchProcess = TaskExecutor()
        # executes all tasks
        batchProcess.execute(taskList)

    except KeyboardInterrupt:
        logger.info("User interrupted")

    except Exception as ex:
        logger.error(ex)
        logger.debug("Trace:\n", exc_info=ex)
        exit(1)

    finally:
        if batchProcess:
            batchProcess.log_summary(logger)
        if closeDriver and driver:
            driver.close()


from .task_base import ComboboxChange, Task
from selenium import webdriver
from selenium.webdriver.support.ui import Select
from typing import List


class ChangeUserClubProperty(ComboboxChange):

    def __init__(self, driver: webdriver, uid: int, flightClub: str, customPropertyNo: int = 1):
        '''
        `ui` user to edit
        `flightClub` custom property value to set (by string representation)
        '''
        self.uid = uid
        url = f"https://vereinsflieger.de/member/community/editfunctions.php?uid={uid}"
        super().__init__(driver=driver,
                         changeUrl=url,
                         cmbxTextValue=flightClub,
                         cmbxHtmlName=f"suc_prop_512_{customPropertyNo}")

    def ident(self) -> str:
        return f"Change user club (uid='{self.uid}') to '{self.cmbxTextValue}'"

    @classmethod
    def parameters(self) -> List[str]:
        return ['uid', 'flightClub', 'customPropertyNo']


class ChangeUserStatus(ComboboxChange):

    def __init__(self, driver: webdriver, uid: int, statusName: str):
        '''
        `ui` user to edit
        `statusName` status to be set
        '''
        self.uid = uid
        url = f"https://vereinsflieger.de/member/community/editcommunity.php?uid={uid}"
        super().__init__(driver=driver,
                         changeUrl=url,
                         cmbxTextValue=statusName,
                         cmbxHtmlName=f"frm_msid")

    def ident(self) -> str:
        return f"Change user status (uid='{self.uid}') to '{self.cmbxTextValue}'"

    @classmethod
    def parameters(self) -> List[str]:
        return ['uid', 'statusName']


class AddCommunityCosts(Task):

    def __init__(self, driver: webdriver, uid: int, costType: str,
                 comment: str, validFrom: str, validTo: str):
        '''
        GUI item: Gebühren -> Neuer Datensatz
        '''
        self.driver = driver
        self.url = f"https://vereinsflieger.de/member/community/editcommunitycost.php?uid={uid}"

        self.uid = uid
        self.costType = costType
        self.comment = comment
        self.validFrom = validFrom
        self.validTo = validTo

    @classmethod
    def parameters(self) -> List[str]:
        return ['uid', 'costType', 'comment', 'validFrom', 'validTo']

    def execute(self) -> bool:
        driver = self.driver

        # open webpage
        driver.get(self.url)

        newCostButton = driver.find_element_by_id('TlbItem1')
        newCostButton.click()

        costType = driver.find_element_by_name('frm_ctid')
        selector = Select(costType)
        selector.select_by_visible_text(self.costType)

        comment = driver.find_element_by_name('frm_comment')
        comment.send_keys(self.comment)

        validFrom = driver.find_element_by_name('frm_validfrom')
        validFrom.clear()
        validFrom.send_keys(self.validFrom)

        validTo = driver.find_element_by_name('frm_validto')
        validTo.send_keys(self.validTo)

        costType.submit()

        # TODO implement a checker for valid input (may make use of JS function: checkMandatory())
        return True

    def ident(self) -> str:
        return f"AddCommunityCosts for user {self.uid}"



from bulk_edit.tasks import Task
from typing import List
from collections import deque
from typing import Optional
from dataclasses import dataclass
import datetime
import time
import logging
logger = logging.getLogger(__name__)


class FinishEstimator:

    def __init__(self, allItems: int, windowSize: int = 5):
        self.statisticWindow = windowSize
        self.reset(allItems)

    def reset(self, allItems: int):
        self.allItems = allItems
        self.loopCounter = 0
        # items back in history for average computation
        initialWindow = [None for i in range(self.statisticWindow)]
        self.windowTickTimes = deque(initialWindow, self.statisticWindow)
        self.timePerItem = None

    def percent(self) -> float:
        return 100.0/float(self.allItems) * float(self.loopCounter)

    def items_left(self):
        return self.allItems - self.loopCounter

    def time_until_complete(self) -> str:
        if self.timePerItem is None:
            return "<estimation in progress>"
        else:
            itemsLeft = self.allItems - self.loopCounter
            return f"{int(itemsLeft * self.timePerItem)} sec"

    def tick(self):
        now = datetime.datetime.now()
        self.loopCounter += 1

        firstTimeInWindow = self.windowTickTimes.popleft()

        if firstTimeInWindow:
            windowDelta = now - firstTimeInWindow
            self.timePerItem = float(windowDelta.seconds) / float(self.statisticWindow)

        self.windowTickTimes.append(now)

    def log_progress(self, logger: logging.Logger):
        message = ("Progress: {percent:.2f} % (Left: {no} items in "
                   "approx. {time})").format(percent=self.percent(),
                                             no=self.items_left(),
                                             time=self.time_until_complete())
        logger.info(message)


@dataclass
class ExceptionalTask:
    task: Task
    exception: Exception


class TaskExecutor:

    def __init__(self):
        self.exceptionalTasks: List[ExceptionalTask] = []
        self.executeCount = 0
        self.numberOfTasks = 0

    def execute(self, tasks: List[Task], delay: Optional[float] = 1.5):
        '''
        Executes `tasks`` one by one.
        Between each task a `delay` in seconds can be awaited. (This is useful,
        because VF will block our IP for a curtain time if we fire to much requests
        within a given time period.)
        '''
        self.exceptionalTasks = []
        self.executeCount = 0
        self.numberOfTasks = len(tasks)
        progress = FinishEstimator(self.numberOfTasks)

        for task in tasks:
            if delay:
                time.sleep(delay)

            try:
                self.executeCount += 1
                task.execute()

            except Exception as ex:
                self.exceptionalTasks.append(ExceptionalTask(task, ex))
                # TODO give the user more info which task fails
                logger.error(f"Exception received while executing a task. Exception: {ex}")
                logger.debug("Trace:\n", exc_info=ex)

            # some statistics
            progress.tick()
            progress.log_progress(logger)

    def log_summary(self, logger: logging.Logger):

        indent = '\t'
        linebreak = '\n'
        message = []
        fails = len(self.exceptionalTasks)

        if fails > 0:
            message.append("The following tasks failed (executed: {a}, failed: {b}, {p:.2f} "
                           "% fail)".format(a=self.executeCount,
                                            b=fails,
                                            p=100.0/self.executeCount*fails))
        else:
            message.append("All tasks succeeded :)")

        for fail in self.exceptionalTasks:
            message.append(linebreak)
            message.append(indent)
            message.append(fail.task.ident())
            # Get the message from exception. (Selenium pushes an line break at each exception
            # representation). This gets the message or string represantion as fallback:
            exceptionMessage = getattr(fail.exception, 'msg', str(fail.exception))
            message.append(f" (Reason: {exceptionMessage})")

        logger.info(''.join(message))



import abc
from selenium import webdriver
from selenium.webdriver.support.ui import Select
from typing import List


class Task:

    @abc.abstractmethod
    def execute(self) -> bool:
        pass

    def ident(self) -> str:
        return type(self).__name__

    @classmethod
    def parameters(cls) -> List[str]:
        return []


class ComboboxChange(Task):

    def __init__(self, driver: webdriver, changeUrl: str, cmbxHtmlName: str,
                 cmbxTextValue: str):
        self.driver = driver
        self.cmbxTextValue = cmbxTextValue
        # this is the html name property for the custom property drop down (on editfunctions.php)
        self.cmbxHtmlName = cmbxHtmlName
        self.changeUrl = changeUrl

        def ident(self) -> str:
            return (f"Swap combobox value to '{self.cmbxTextValue}' (url={self.changeUrl} "
                    f"comboboxName={self.cmbxHtmlName})")

    @classmethod
    def parameters(cls) -> List[str]:
        return ['changeUrl', 'cmbxHtmlName', 'cmbxTextValue']

    def execute(self) -> bool:
        driver = self.driver

        # open webpage
        driver.get(self.changeUrl)

        # get dropdown element and select value
        combobox = driver.find_element_by_name(self.cmbxHtmlName)
        selector = Select(combobox)
        selector.select_by_visible_text(self.cmbxTextValue)

        # hit the submit method on element. (This lets silenumn walk up the DOM tree
        # to the <form> tag and submit it.)
        combobox.submit()

        # TODO implement a checker for valid input (may make use of JS function: checkMandatory())
        return True


def print_subclass(task_t: type, name_filter: str = '*'):
    for derivedTask in task_t.__subclasses__():
        if name_filter == '*' or name_filter.lower() in derivedTask.__name__.lower():
            # print the task and parameters
            taskName = '{name}\n'
            taskParameter = '  * {parameterName}\n'

            print(taskName.format(name=derivedTask.__name__))
            parameterList = [taskParameter.format(parameterName=p)
                            for p in derivedTask.parameters()]
            print(''.join(parameterList))
        # continue with subclasses (if there are any)
        if len(derivedTask.__subclasses__()) > 0:
            print_subclass(derivedTask, name_filter)


import os
import sys

from setuptools import setup


name = 'Jan Gietzel'
mail = 'jan.gietzel@gmail.com'
__author__ = f'{name} <{mail}>'


with open('readme.md') as readme_file:
    long_description = readme_file.read()


# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()
elif sys.argv[-1] == 'clean':
    import shutil
    if os.path.isdir('build'):
        shutil.rmtree('build')
    if os.path.isdir('dist'):
        shutil.rmtree('dist')
    if os.path.isdir('vf_utils.egg-info'):
        shutil.rmtree('vf_utils.egg-info')


setup(
    name="vf_utils",
    version="0.1.1",
    author=name,
    author_email=mail,
    description="Doing bulk edit jobs for vereinsflieger.de",
    license="MIT",
    keywords="vereinsflieger.de",
    url="https://github.com/Ka55i0peia/vf_utils",
    packages=['bulk_edit', 'bulk_edit.tasks'],
    long_description_content_type='text/markdown',
    long_description=long_description,
    python_requires='>=3',
    install_requires=[
          'selenium>=3.141.0',
          'argparse>=1.4.0'
    ],
    setup_requires=["wheel"]
)


from bulk_edit.tasks import Task
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from typing import List, Tuple
from getpass import getpass
import logging
logger = logging.getLogger(__name__)


class LoginToVf(Task):

    def __init__(self, driver: webdriver, credentialFile: str = '', user: str = ''):
        self.driver = driver
        self.credentialFile = credentialFile
        self.user = user

    @classmethod
    def parameters(cls) -> List[str]:
        # TODO add some more info about parameter like type and description
        return ['credentialFile']

    def ident(self) -> str:
        return f"VF Login with credentialFile '{self.credentialFile}'"

    # NOTE candidate for global utility
    @staticmethod
    def get_user_and_passwd_from_file(fileName: str) -> Tuple[str, str]:
        user = None
        pwd = None
        with open(fileName, 'r') as f:
            # first line:
            user = f.readline()
            # second line:
            pwd = f.readline()

        # error check
        if len(user) == 0 or len(pwd) == 0:
            raise Exception(f"Crendential file '{fileName}' must contain a user "
                            "in the first line and password in the second line")

        return (user, pwd)

    @staticmethod
    def get_user_and_pwd_from_cli(user: str = '') -> Tuple[str, str]:
        print('Please enter user password to login to vereinflieger.de')
        if user == '':
            user = input('user: ')
        pwd = getpass('password: ')
        return (user, pwd)

    def execute(self) -> bool:
        self.driver.get("https://vereinsflieger.de/")

        # check if we are already logged in ..
        try:
            self.driver.find_element_by_link_text("Mein Profil")
            return True
        except NoSuchElementException:
            # no, we aren't logged in already
            pass

        # get credentials
        if self.credentialFile is not None:
            try:
                (user, pwd) = type(self).get_user_and_passwd_from_file(self.credentialFile)
            except Exception as ex:
                logger.error(f"Reading credential file '{self.credentialFile}' failed.")
                logger.debug(ex, exc_info=ex)
                # fallback
                (user, pwd) = type(self).get_user_and_pwd_from_cli(self.user)
        else:
            (user, pwd) = type(self).get_user_and_pwd_from_cli(self.user)

        # login
        self.driver.find_element_by_name("user").send_keys(user)
        self.driver.find_element_by_name("pwinput").send_keys(pwd)
        self.driver.find_element_by_name("submit").click()

        return True
